#include "debugger.h"

DebugStatus_t DebuggerStatus[dbg_cntElements];

String IpAddress2String(const IPAddress &ipAddress);
void processCmdDebug(String command);
void Debug_formatCFG();
void Debug_formatPersistence();
void Debug_printSystemInfo();
void Debug_printWifiInfo();
void Debug_CheckEEPOM();
void Debug_dumpConfig();
void Debug_dumpPersistance();
void Debug_ShowDTCs();
void Debug_dumpGlobals();
void Debug_printHelp();

void initDebugger()
{
    DebuggerStatus[dbg_Serial] = disabled;
    DebuggerStatus[dbg_Webui] = disabled;

    Serial.setDebugOutput(false);
}

void Debug_Process()
{
    typedef enum InputProcessed_e
    {
        IDLE,
        CMD_COMPLETE,
        CMD_ABORT,
        CMD_OVERFLOW
    } InputProcessed_t;

    static unsigned int inputCnt = 0;
    static char inputBuffer[32];
    InputProcessed_t InputProcessed = IDLE;

    if (Serial.available())
    {
        char inputChar = Serial.read();

        switch (inputChar)
        {
        case '\n':
            inputBuffer[inputCnt] = 0; // terminate the String
            inputCnt = 0;
            InputProcessed = CMD_COMPLETE;
            break;

        case 0x1B: // Esc
            inputBuffer[0] = 0;
            inputCnt = 0;
            InputProcessed = CMD_ABORT;
            break;

        case 0x21 ... 0x7E: // its a real letter or sign and not some control-chars
            inputBuffer[inputCnt] = inputChar;
            inputCnt++;
            break;

        default:
            break;
        }

        if (inputCnt > sizeof(inputBuffer))
        {
            inputCnt = 0;
            inputBuffer[sizeof(inputBuffer) - 1] = 0; // terminate the String
            InputProcessed = CMD_OVERFLOW;
        }
    }

    switch (InputProcessed)
    {
    case CMD_ABORT:
        Debug_pushMessage("Abort\n");
        break;

    case CMD_COMPLETE:
        processCmdDebug(String(inputBuffer));
        break;

    case CMD_OVERFLOW:
        Debug_pushMessage("input Buffer overflow\n");
        break;

    default:
        break;
    }
    InputProcessed = IDLE;
}

void SetDebugportStatus(DebugPorts_t port, DebugStatus_t status)
{
    if (status == disabled)
        Debug_pushMessage("disable DebugPort %s\n", sDebugPorts[port]);

    DebuggerStatus[port] = status;

    if (status == enabled)
        Debug_pushMessage("enabled DebugPort %s\n", sDebugPorts[port]);
}

void Debug_pushMessage(const char *format, ...)
{
    if ((DebuggerStatus[dbg_Serial] == enabled) || (DebuggerStatus[dbg_Webui] == enabled))
    {
        char buff[64];
        va_list arg;
        va_start(arg, format);
        vsnprintf(buff, sizeof(buff), format, arg);
        va_end(arg);

        if (DebuggerStatus[dbg_Serial] == enabled)
        {
            Serial.print(buff);
        }
        if (DebuggerStatus[dbg_Webui] == enabled)
        {
            Websocket_PushLiveDebug(String(buff));
        }
    }
}

void processCmdDebug(String command)
{
    if (command == "help")
        Debug_printHelp();
    else if (command == "sysinfo")
        Debug_printSystemInfo();
    else if (command == "netinfo")
        Debug_printWifiInfo();
    else if (command == "formatCFG")
        Debug_formatCFG();
    else if (command == "formatPDS")
        Debug_formatPersistence();
    else if (command == "checkEE")
        Debug_CheckEEPOM();
    else if (command == "dumpEE1k")
        dumpEEPROM(0, 1024);
    else if (command == "dumpEE")
        dumpEEPROM(0, EEPROM_SIZE_BYTES);
    else if (command == "resetPageEE")
        MovePersistencePage_EEPROM(true);
    else if (command == "dumpCFG")
        Debug_dumpConfig();
    else if (command == "dumpPDS")
        Debug_dumpPersistance();
    else if (command == "saveEE")
        globals.requestEEAction = EE_ALL_SAVE;
    else if (command == "showdtc")
        Debug_ShowDTCs();
    else if (command == "dumpGlobals")
        Debug_dumpGlobals();
    else if (command == "sdbg")
        SetDebugportStatus(dbg_Serial, enabled);
    else
        Debug_pushMessage("unknown Command\n");
}

void Debug_formatCFG()
{
    Debug_pushMessage("Formatting Config-EEPROM and reseting to default\n");
    FormatConfig_EEPROM();
}

void Debug_formatPersistence()
{
    Debug_pushMessage("Formatting Persistence-EEPROM and reseting to default\n");
    FormatPersistence_EEPROM();
}

void Debug_printSystemInfo()
{
    Debug_pushMessage("Hostname: %s\n", globals.DeviceName);

    FlashMode_t ideMode = ESP.getFlashChipMode();
    Debug_pushMessage("Sdk version: %s\n", ESP.getSdkVersion());
    Debug_pushMessage("Core Version: %s\n", ESP.getCoreVersion().c_str());
    Debug_pushMessage("Boot Version: %u\n", ESP.getBootVersion());
    Debug_pushMessage("Boot Mode: %u\n", ESP.getBootMode());
    Debug_pushMessage("CPU Frequency: %u MHz\n", ESP.getCpuFreqMHz());
    Debug_pushMessage("Reset reason: %s\n", ESP.getResetReason().c_str());
    Debug_pushMessage("Flash Size: %d\n", ESP.getFlashChipRealSize());
    Debug_pushMessage("Flash Size IDE: %d\n", ESP.getFlashChipSize());
    Debug_pushMessage("Flash ide mode:  %s\n", (ideMode == FM_QIO ? "QIO" : ideMode == FM_QOUT ? "QOUT"
                                                                        : ideMode == FM_DIO    ? "DIO"
                                                                        : ideMode == FM_DOUT   ? "DOUT"
                                                                                               : "UNKNOWN"));
    Debug_pushMessage("OTA-Pass: %s\n", QUOTE(ADMIN_PASSWORD));
    Debug_pushMessage("Git-Revison: %s\n", constants.GitHash);
    Debug_pushMessage("Sw-Version: %d.%02d\n", constants.FW_Version_major, constants.FW_Version_minor);
}

void Debug_dumpConfig()
{
    Debug_pushMessage("batteryType: %d\n", ConfigData.batteryType);
    Debug_pushMessage("EEPROM_Version: %d\n", ConfigData.EEPROM_Version);
    Debug_pushMessage("checksum: 0x%08X\n", ConfigData.checksum);
}

void Debug_dumpGlobals()
{
    Debug_pushMessage("systemStatus: %d\n", globals.systemStatus);
    Debug_pushMessage("battery_level: %d\n", globals.battery_level);
    Debug_pushMessage("loadvoltage_mV: %d\n", globals.loadvoltage_mV);
    Debug_pushMessage("requestEEAction: %d\n", globals.requestEEAction);
    Debug_pushMessage("DeviceName: %s\n", globals.DeviceName);
    Debug_pushMessage("DeviceName_ID: %s\n", globals.DeviceName_ID);
    Debug_pushMessage("FlashVersion: %s\n", globals.FlashVersion);
    Debug_pushMessage("eePersistanceAdress: %d\n", globals.eePersistanceAdress);
    Debug_pushMessage("hasDTC: %d\n", globals.hasDTC);
}

void Debug_dumpPersistance()
{
    Debug_pushMessage("writeCycleCounter: %d\n", PersistenceData.writeCycleCounter);
    Debug_pushMessage("activeFaction: %d\n", PersistenceData.activeFaction);
    Debug_pushMessage("faction_1_timer: %d\n", PersistenceData.faction_1_timer);
    Debug_pushMessage("faction_2_timer: %d\n", PersistenceData.faction_2_timer);
    Debug_pushMessage("faction_3_timer: %d\n", PersistenceData.faction_3_timer);
    Debug_pushMessage("checksum: %d\n", PersistenceData.checksum);
    Debug_pushMessage("PSD Adress: 0x%04X\n", globals.eePersistanceAdress);
}

void Debug_printWifiInfo()
{
}

void Debug_CheckEEPOM()
{
    uint32_t checksum = PersistenceData.checksum;
    PersistenceData.checksum = 0;

    if (Checksum_EEPROM((uint8_t *)&PersistenceData, sizeof(PersistenceData)) == checksum)
    {
        Debug_pushMessage("PersistenceData EEPROM Checksum OK\n");
    }
    else
    {
        Debug_pushMessage("PersistenceData EEPROM Checksum BAD\n");
    }

    PersistenceData.checksum = checksum;

    checksum = ConfigData.checksum;
    ConfigData.checksum = 0;

    if (Checksum_EEPROM((uint8_t *)&ConfigData, sizeof(ConfigData)) == checksum)
    {
        Debug_pushMessage("ConfigData EEPROM Checksum OK\n");
    }
    else
    {
        Debug_pushMessage("ConfigData EEPROM Checksum BAD\n");
    }
    ConfigData.checksum = checksum;
}

void Debug_ShowDTCs()
{
    char buff_timestamp[16]; // Format: DD-hh:mm:ss:xxx
    char buff_active[9];

    Debug_pushMessage("\n      timestamp | DTC-Nr. |   status | severity | debugVal\n");

    for (uint32_t i = 0; i < MAX_DTC_STORAGE; i++)
    {
        if (DTCStorage[i].Number < DTC_LAST_DTC)
        {
            sprintf(buff_timestamp, "%02d-%02d:%02d:%02d:%03d",
                    DTCStorage[i].timestamp / 86400000,    // Days
                    DTCStorage[i].timestamp / 360000 % 24, // Hours
                    DTCStorage[i].timestamp / 60000 % 60,  // Minutes
                    DTCStorage[i].timestamp / 1000 % 60,   // Seconds
                    DTCStorage[i].timestamp % 1000);       // milliseconds

            if (DTCStorage[i].active == DTC_ACTIVE)
                strcpy(buff_active, "active");
            else if (DTCStorage[i].active == DTC_PREVIOUS)
                strcpy(buff_active, "previous");
            else
                strcpy(buff_active, "none");

            Debug_pushMessage("%s   %7d   %8s   %8d   %8d\n", buff_timestamp, DTCStorage[i].Number, buff_active, DTCStorage[i].severity, DTCStorage[i].debugVal);
        }
    }
}

void Debug_printHelp()
{
    char buff[64];

    for (int i = sizeof(helpCmd) / 63; i < sizeof(helpCmd) / 63; i++)
    {
        memcpy_P(buff, (helpCmd + (i * 63)), 63);
        buff[63] = 0;
        Debug_pushMessage(buff);
    }
}